/*======================================================

	SD6031/WIF Firmware
	SDHC / FAT32 support version
	
	2016.3.28 by Koichi Nishida
	Copyright (C) 2016 Koichi NISHIDA & Esubi
	email : tulip-house@msf.biglobe.ne.jp

======================================================*/

/*
hardware information of SD6031
use ATMEGA164P AVR.
connect 16MHz crystal with the AVR.

the fuse setting :
(note that JTAG must be disable by this setting)

 LOW:  11011110
 HIGH: 11011001

port direction and connection

DDRB = SD card (CLK,DO,DI,CS,EJECT,WP),LED2,LED1
DDRA = 0b11111111; data out to P6
DDRC = 0b00000000; data in from P6
DDRD = 0b11110000; ctrl out to P6 (ATN,DAC,RFD,DAV), ctrl in from P6 (ATN,DAC,RFD,DAV)
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <string.h>
#include "SD.h"
#include <util/delay.h>

#define WAIT 1

#define min(a,b) ((a)>(b)?(b):(a))

// ==================== prototypes ====================

// calculate 2^x
uint8_t pow2(register uint8_t x);

#define DRIVE_NUM 4

// send P31 image file names to P6
void sendFileNames(uint8_t drive);
	
// ---------- communication with P6  ----------
// communicate with PC-6001
// send data to PC-6001
void P6_snd(uint8_t d1, uint8_t d2);
// receive data to PC-6001
void P6_rcv(uint8_t *d1, uint8_t *d2);

// send double data to PC-6001
void P6_sndw_pre(void);
void P6_sndw(uint8_t d1, uint8_t d2, uint8_t d3, uint8_t d4);
void P6_sndw_post(void);
// receive double data to PC-6001
void P6_rcvw_pre(void);
void P6_rcvw(uint8_t *d1, uint8_t *d2, uint8_t *d3, uint8_t *d4);
void P6_rcvw_post(void);

// ---------- PC-6031 status ----------
struct pc6031_status {
	uint8_t drive;
	uint16_t track;
	uint8_t sector;
	uint8_t length;
	uint8_t valid;
	uint8_t mode;
	uint8_t startSec[DRIVE_NUM];
	uint8_t endSec[DRIVE_NUM];
	uint8_t error;
};

// initialize PC-6031 status
void pc6031_init(void);

// process PC-6031 command
void pc6031_process(void);

//==================== implementation ====================

struct FILE image[DRIVE_NUM];

struct pc6031_status pc6031;	// PC-6031 status
uint8_t LED;				// LED status
uint8_t RST_init;

// calculate 2^x
uint8_t pow2(register uint8_t x)
{
	uint8_t i, r=1;

	for (i=0; i!=x; i++) {
		r<<=1;
	}
	return r;
}

void sendAFileName(struct FILELST *list, uint16_t *entryNum, uint32_t blkadr, uint16_t ofs)
{
	uint8_t err = 0;
	
	SD_readBlock(blkadr, &err);
	if (err) return;
	P6_snd(1, 0);
	for (uint8_t i=0; i<8; i++) P6_snd(SD_p.buff[ofs+i],0);
}

// send P31 image file names to P6
void sendFileNames(uint8_t drive)
{
	uint8_t err = 0;	
		
	if (bit_is_set(PINB,3)) return;
	
	struct FILE directory;
	char ext_[4]="   ", name_[9]="D       ";
	name_[1] = ('0'+drive);
	FILE_open(&directory, 0, name_, ext_, &err);
	if (err) return;
	struct FILELST list;
	uint16_t entryNum;
	
	FILE_traverseADirectory(directory.startCluster, "P31", &list, &entryNum, &sendAFileName, &err);
	if (err) return;

	P6_snd(0, 0);
}

// called when the card is inserted or removed
void SD_checkEject(void)
{
	uint32_t i;
	uint8_t j,k;
	uint8_t wt;

	if (bit_is_set(PINB,3)) {
		for (j=0; j<DRIVE_NUM; j++) {
			image[j].valid = 0;
		}
		SD_p.inited = 0;
	} else if (!SD_p.inited) {
		wt = (RST_init) ? 3 : 6;
		for (k=0; k!=wt;k++)

		for (i=0; i!=0x50000;i++)
			if (bit_is_set(PINB,3)) return;
			
		if (RST_init == 1) {
			LED = 0b00000011;
			PORTB = ((PORTB&0b11111100) | LED);
		} else {
			LED = 0b00000001;
			PORTB = ((PORTB&0b11111100) | LED);
		}
		uint8_t err = 0;

		SD_init(&err);
		
		
		LED = 0b00000000;
		PORTB = ((PORTB&0b11111100) | LED);
	}
}

// communicate with PC-6001
// send data to PC-6001
void P6_snd(uint8_t d1, uint8_t d2)
{ 
	while (bit_is_clear(PIND,1)) ;	// wait RFD set
	PORTA = d1;
	PORTD |= 0b00010000;			// set DAV
	while (bit_is_clear(PIND,2)) ;	// wait DAC set
	PORTA = d2;
	PORTD &= 0b11101111;			// reset DAV
	while (bit_is_set(PIND,2)) ;	// wait DAC reset
}

// receive data from PC-6001
void P6_rcv(uint8_t *d1, uint8_t *d2)
{
	*d1 = 0xff;
	*d2 = 0xff;
	PORTD |= 0b00100000;								// set RFD
	while (bit_is_clear(PIND,0)) ;						// wait DAV set
	PORTD &= 0b11011111;								// reset RFD
	*d1 = PINC;
	PORTD |= 0b01000000;								// set DAC
	while (bit_is_set(PIND,0)) ;						// wait DAV reset
	if (d2) *d2 = PINC;
	PORTD &= 0b10111111;								// reset DAC
}

// pre-send double data to PC-6001
void P6_sndw_pre(void)
{ 
	while (bit_is_clear(PIND,1)) ;	// wait RFD set
	DDRA  = 0b11111111;				// data out to P6
	DDRC  = 0b11111111;				// data out to P6
}

// send double data to PC-6001
void P6_sndw(uint8_t d1, uint8_t d2, uint8_t d3, uint8_t d4)
{ 
	PORTA = d1;
	PORTC = d2;
	PORTD |= 0b00010000;			// set DAV
	while (bit_is_clear(PIND,2)) ;	// wait DAC set
	PORTA = d3;
	PORTC = d4;
	PORTD &= 0b11101111;			// reset DAV
	while (bit_is_set(PIND,2)) ;	// wait DAC reset
}

// post-send double data to PC-6001
void P6_sndw_post(void)
{ 
	PORTC = 0b11111111;
	DDRA  = 0b11111111;	// data out to P6
	DDRC  = 0b00000000;	// data in from P6
}

// pre-receive double data to PC-6001
void P6_rcvw_pre(void)
{ 
	PORTA = 0b11111111;
	PORTC = 0b11111111;
	DDRA  = 0b00000000;	// data in from P6
	DDRC  = 0b00000000;	// data in from P6
	PORTD |= 0b00100000;								// set RFD
	while (bit_is_clear(PIND,0)) ;						// wait DAV set
	PORTD &= 0b11011111;								// reset RFD
}

// receive double data from PC-6001
void P6_rcvw(uint8_t *d1, uint8_t *d2, uint8_t *d3, uint8_t *d4)
{
	while (bit_is_clear(PIND,0)) ;						// wait DAV set
	*d1 = PINC;
	*d2 = PINA;
	PORTD |= 0b01000000;								// set DAC
	while (bit_is_set(PIND,0)) ;						// wait DAV reset
	*d3 = PINC;
	*d4 = PINA;
	PORTD &= 0b10111111;								// reset DAC
}

// post-receive double data to PC-6001
void P6_rcvw_post(void)
{ 
	PORTC = 0b11111111;
	DDRA  = 0b11111111;	// data out to P6
	DDRC  = 0b00000000;	// data in from P6
}

// PC interrupt is used for detecting SD card eject
ISR(PCINT1_vect)
{
	SD_p.inited = 0;
	SD_checkEject();
}

// initialize PC-6031 status
void pc6031_init(void)
{
	uint16_t i;

	pc6031.valid = 0;
	pc6031.error = 0;
	pc6031.mode = 0;
	for (i = 0;i != DRIVE_NUM;i ++) {
		pc6031.startSec[i] = 1;
		pc6031.endSec[i]   = 16;
	}
}

// process PC-6031 command
void pc6031_process(void)
{
	while (1) {
		uint8_t cmd;

		while (bit_is_clear(PIND,3)) ;	// wait attention
		P6_rcv(&cmd, 0);				// receive command
		while (bit_is_set(PIND,3)) ;	// wait not attention
		switch (cmd) {
		case 0:		// init
			pc6031.error = 0;	
			break;
		case 1:			// write
		case 17: 		// high speed write
		case 0x83:		// expand write
		case 0x84: 		// expand high speed write
		case 0x86: 		// double high speed write
		case 0x87: {	// expand double high speed write
			uint8_t lg, dr, trkupper,trklower, sec, i;
			uint16_t j;
			uint16_t trk;
			
			pc6031.error = 0;
			trkupper = 0;
			P6_rcv(&lg, 0); P6_rcv(&dr, 0);
			if (cmd==0x83 || cmd==0x84 || cmd==0x87) P6_rcv(&trkupper, 0);
			P6_rcv(&trklower, 0);
			trk  = (uint16_t)trklower;
			trk += (uint16_t)trkupper*0x100;
			P6_rcv(&sec, 0);

			if (!image[dr].startCluster) { pc6031.error = 1; break; }
			LED = 0b00000010;
			PORTB = ((PORTB&0b11111100) | LED);	
			if (!(pow2(dr)&pc6031.mode)) trk<<=1;
			i = 0;
			do {
				if (image[dr].protect) {
					pc6031.error = 1;
				} else {	
					uint8_t err=0;
					uint32_t long_sector  = (uint32_t)(sec-pc6031.startSec[dr]);
					long_sector += (uint32_t)trk * (uint32_t)(pc6031.endSec[dr]-pc6031.startSec[dr]+1);

					FILE_writeBegin(&image[dr], long_sector, &err);
					if (err) {
						pc6031.error = 1;
					}
				}
				if (cmd==1 || cmd==0x83)
					for (j=0; j!=256; j++) {
						uint8_t err = 0;
						uint8_t d;
						
						P6_rcv(&d, 0);
						if (pc6031.error == 0) SPI_writeByte(d, &err);
					}
				else if (cmd==17 || cmd==0x84)
					for (j=0; j!=128; j++) {
						uint8_t err = 0;
						uint8_t d1, d2;
						
						P6_rcv(&d1, &d2);
						if (pc6031.error == 0) {
							SPI_writeByte(d1, &err);
							SPI_writeByte(d2, &err);
						}
					}			
				else {
					P6_rcvw_pre();
					for (j=0; j!=64; j++) {
						uint8_t d1, d2, d3, d4;
						
						P6_rcvw(&d1, &d2, &d3, &d4);
						if (pc6031.error == 0) {
							uint8_t err = 0;
							SPI_writeByte(d1, &err);
							SPI_writeByte(d2, &err);
							SPI_writeByte(d3, &err);
							SPI_writeByte(d4, &err);
						}
					}			
					P6_rcvw_post();
				}

				if (pc6031.error == 0) {
					uint8_t err = 0;
					for (j=0; j!=256; j++) SPI_writeByte(0, &err);
					FILE_writeEnd(&err);
				}
				if (sec == pc6031.endSec[dr]) {
					sec = pc6031.startSec[dr];
				} else {
					sec++;
				}
				i ++;
			} while (i != lg);
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		}
		case 2:			// read
		case 0x85: {	// expand read
			uint8_t trkupper,trklower;
			trkupper = 0;
			P6_rcv(&pc6031.length, 0);
			P6_rcv(&pc6031.drive, 0);
			if (cmd==0x85) P6_rcv(&trkupper, 0);
			P6_rcv(&trklower, 0);
			pc6031.track  = (uint16_t)trklower;
			pc6031.track += (uint16_t)trkupper*0x100;
			P6_rcv(&pc6031.sector, 0);
			
			if (!image[pc6031.drive].valid) { pc6031.error=1; break; }
			LED = 0b00000001;
			PORTB = ((PORTB&0b11111100) | LED);
			if (!(pow2(pc6031.drive)&pc6031.mode)) pc6031.track<<=1;
			pc6031.valid = 1;
			pc6031.error = 0;
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		}
		case 3:			// transfer read data
		case 18: 		// high speed transfer read data
		case 0x88: {	// double high speed transfer read data
			uint8_t i;
			uint16_t j;

			pc6031.error = 0;
			if ((!pc6031.valid) || (!image[pc6031.drive].startCluster)) { pc6031.error=1; break; }
			LED = 0b00000001;
			PORTB = ((PORTB&0b11111100) | LED);
			i = 0;
			do {
				uint8_t err = 0;
				uint32_t long_sector  = (uint32_t)(pc6031.sector-pc6031.startSec[pc6031.drive]);
				long_sector += (uint32_t)pc6031.track * (uint32_t)(pc6031.endSec[pc6031.drive]-pc6031.startSec[pc6031.drive]+1);

				FILE_readBegin(&image[pc6031.drive], long_sector, &err);

				if (err) {
					pc6031.error = 1;
				}
				
				if (cmd==3) for (j=0; j!=256; j++) {
					uint8_t d;
					if (pc6031.error == 0) {
						d = SPI_readByte(&err);
					} else {
						d = 0xff;
					}
					
					P6_snd(d, 0);
				} else if (cmd==18) for (j=0; j!=128; j++) {
					uint8_t d1;
					uint8_t d2;
					if (pc6031.error == 0) {
						d1 = SPI_readByte(&err);
						d2 = SPI_readByte(&err);
					} else {
						d1 = 0xff;
						d2 = 0xff;
					}
					
					P6_snd(d1, d2);
				} else {
					P6_sndw_pre();
					for (j=0; j!=64; j++) {
						uint8_t d1;
						uint8_t d2;
						uint8_t d3;
						uint8_t d4;
						if (pc6031.error == 0) {
							d1 = SPI_readByte(&err);
							d2 = SPI_readByte(&err);
							d3 = SPI_readByte(&err);
							d4 = SPI_readByte(&err);
						} else {
							d1 = 0xff;
							d2 = 0xff;
							d3 = 0xff;
							d4 = 0xff;
						}
						P6_sndw(d1, d2, d3, d4);
					}
					P6_sndw_post();
				}
				if (pc6031.error == 0) {
					for (j=0; j!=256; j++) SPI_readByte(&err);
					FILE_readEnd(&err);
				}
				if (pc6031.sector == pc6031.endSec[pc6031.drive]) {
					pc6031.sector = pc6031.startSec[pc6031.drive];
				} else {
					pc6031.sector++;
				}
				i ++;
			} while (i != pc6031.length);
			pc6031.valid = 0;
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		}
		case 5:	{	// format
			uint8_t trk, sec;
			uint8_t drv, is1d;
			uint16_t j;
				
			pc6031.error = 0;
			P6_rcv(&drv, 0);	
			if (image[drv].protect || !image[drv].startCluster) { pc6031.error = 1; break; }
			is1d = (!(pow2(drv)&pc6031.mode));
			LED = 0b00000010;
			PORTB = ((PORTB&0b11111100) | LED);
			// initialize the area with 0xff
			for (trk=0; trk!=80; trk++) {
				if (is1d && (trk&1) ) continue;
				for (sec=1; sec!=17; sec++) {
					uint8_t err=0;
					uint32_t long_sector  = (uint32_t)(sec-pc6031.startSec[drv]);
					long_sector += (uint32_t)trk * (uint32_t)(pc6031.endSec[drv]-pc6031.startSec[drv]+1);
					FILE_writeBegin(&image[drv], long_sector, &err);
					
					if (err) {
						pc6031.error = 1;
						break;
					}

					// format with 0xff
					for (j=0; j!=256; j++) SPI_writeByte(0xff, &err);
					// dummy 256 bytes
					for (j=0; j!=256; j++) SPI_writeByte(0, &err);
					FILE_writeEnd(&err);
					if (err) {
						pc6031.error = 1;
						break;
					}
				}
				if (pc6031.error == 1) {
					break;
				}
			}
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		}
		case 6:		// command status
			P6_snd(0b10000000 | (pc6031.valid<<6) | pc6031.error, 0);
			pc6031.error = 0;
			break;
		case 7: {	// drive status
			const uint8_t status =
				(DRIVE_NUM==1)?0b00011111:((DRIVE_NUM==2)?0b00111111:
				((DRIVE_NUM==3)?0b01111111:0b11111111));

			P6_snd(status, 0);
			pc6031.error = 0;
			break;
		}
		case 11: {	// transfer FDD's memory data to PC-6001
					// only for address 7EFh (return EFh) & 75Fh (return 77h)
			uint16_t i;
			uint8_t AH, AL, DH, DL;

			P6_rcv(&AH, 0); P6_rcv(&AL, 0); P6_rcv(&DH, 0); P6_rcv(&DL, 0);
			P6_snd((AH==7)?((AL==0x5f)?0x77:((AL==0xef)?0xef:0xff)):0xff, 0);
			for (i = 1; i < ((((short)DH)<<8)|DL); i++) P6_snd(0xff, 0);
			pc6031.error = 0;
			break;
		}
		case 20: {
			uint8_t drv;
			
			P6_rcv(&drv, 0);
			P6_snd(image[drv].protect?0b01000000:0,0);
			pc6031.error = 0;
			break;
		}
		case 23: {	// set mode (1D or 2D)
			uint8_t d;
			P6_rcv(&d, 0);
			pc6031.mode =(d&0b00001111);
			pc6031.error = 0;
			break;
		}
		case 24: {	// ask mode
			P6_snd(pc6031.mode, 0);
			pc6031.error = 0;
			break;
		}
		case 0x80: {	// set SECTOR data
			uint8_t drv;
			P6_rcv(&drv, 0);
			P6_rcv(&(pc6031.startSec[drv]), 0);
			P6_rcv(&(pc6031.endSec[drv])  , 0);
			pc6031.error = 0;
			break;
		}
		case 0x81: {	// ask SECTOR data
			uint8_t drv;
			P6_rcv(&drv, 0);
			P6_snd(pc6031.startSec[drv], 0);
			P6_snd(pc6031.endSec[drv]  , 0);
			pc6031.error = 0;
			break;
		}
		case 0x82: {	// report DISK IMAGE length
			uint8_t drv;
			uint8_t tmp;
			uint32_t tmplen;

			P6_rcv(&drv, 0);
			tmplen = image[drv].length;
			tmp = tmplen / 0x1000000;
			P6_snd(tmp,0);
			tmplen = tmplen - tmp * 0x1000000;
			tmp = tmplen / 0x10000;
			P6_snd(tmp,0);
			tmplen = tmplen - tmp * 0x10000;
			tmp = tmplen / 0x100;
			P6_snd(tmp,0);
			tmplen = tmplen - tmp * 0x100;
			tmp = tmplen;
			P6_snd(tmp,0);

			pc6031.error = 0;
			break;
		}
		case 0xFD: { // send P31 image names to P6
			uint8_t drv;

			P6_rcv(&drv, 0);
			LED = 0b00000001;
			PORTB = ((PORTB&0b11111100) | LED);
			sendFileNames(drv);
			pc6031.error = 0;
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		}
		case 0xFE: { // change image
			uint8_t drv, i;
			char name[23]="D                  P31";

			pc6031.error = 1;
			P6_rcv(&drv, 0);
			name[1] = ('0'+drv);
			LED = 0b00000001;
			PORTB = ((PORTB&0b11111100) | LED);
			for (i=0; i!=8; i++)
				P6_rcv((uint8_t *)&(name[i+11]), 0);
			uint8_t err = 0;
			FILE_openAbs(&image[drv], name, &err);
			if (!err) {
				pc6031.error = 0;
				
				struct FILE tempImg;
				memcpy(&name[11],"2D         ",11);
				err = 0;
				FILE_openAbs(&tempImg, name, &err);
				if (!err) pc6031.mode |= pow2(drv);
			}
			LED = 0;
			PORTB = ((PORTB&0b11111100) | LED);
			break;
		}}
	}
}

int main()
{	
	_delay_ms(100);
	
	
	//PORTB &= 0b11111100;
	PORTA = 0b00000000;
	PORTC = 0b11111111;
	PORTD = 0b00001111;	// ATN output is always 0

	DDRB |= 0b00000011; // LED2,LED1
	DDRA = 0b11111111;	// data out to P6
	DDRC = 0b00000000;	// data in from P6
	DDRD = 0b11110000;	// ctrl out to P6 (ATN,DAC,RFD,DAV),
						// ctrl in from P6 (ATN,DAC,RFD,DAV)

	uint8_t err = 0;
	SD_init(&err);
	
	// PC int is used for SD card's eject
	PCMSK1 |= 0x08;
	PCICR |= (1<<PCIE1);

	RST_init = 1;
	LED = 0;
	PORTB = ((PORTB&0b11111100) | LED);

	pc6031_init();
	
	SD_p.inited = 0;
		
	SD_checkEject();
	
	RST_init = 0;

	sei();
	
	pc6031_process();
}
